
#ifndef ENABLE_FILE
#error KEYBOARD: requires file support!
#endif

#include <stdtypes.h>

#include <keyboard.h>
#include <kalloc.h>
#include <asm.h>
#include <idt.h>
#include <video.h>
#include <sys.h>
#include <device.h>
#include <ascii.h>

/*///////////////////////////////////////////////////////////////// */
/* Keyqueue */

/*
	These functions provide a queue for keyboard input, so
	that the key presses do not need to be handled immediately.
	
	This is only the queue management stuff, the actual work is
	done in keyboard.c
	
	We translate the scancodes before queuing (sp?) them,
	so only ASCII characters will be put in the queue. All
	actual commands are handled immediately.
	
	The queue is placed in memory at 0x5000, after the device
	table, and is 512 bytes long (so we can queue up to 512
	characters, remembering that some keys generate more than
	one character).
	
*/

#define KEY_QUEUE_SIZE	(0x200)
#define KEY_QUEUE_MASK	(0x1FF)
#define NEXT(x) ((x+1) & KEY_QUEUE_MASK)
ulong_t qHead, qTail;
word_t keyState;		
char_t queue_ch;

/*
	qHead is inclusive, qTail is not.
	ie. There is no data at qTail.
	
	We must leave a gap at the end of the table when qTail == qHead - 1,
	so that qTail == qHead is unambiguous.
*/

void FlushKeyQueue()
{
	memset((void*)KEY_QUEUE_START, (byte_t)0, KEY_QUEUE_SIZE);
	qHead = qTail = 0;
	keyState=0;
}

void KeyQueueInit()
{
	FlushKeyQueue();
}

int_t EnqueueKey(char_t ch)
{
	/* check if queue is full */
	if (NEXT(qTail) == qHead)
		return -1;
	
	char_t* p = (char_t*)(KEY_QUEUE_START + qTail);
	*p = ch;
	qTail = NEXT(qTail);
	return 0;
}

int_t DequeueKey(char_t* ch)
{
	/* check if queue is empty */
	if (qHead == qTail)
		{return -1;}
	
	char_t* p = (char_t*)(KEY_QUEUE_START + qHead);
	*ch = *p;
	qHead = NEXT(qHead);
	return 0;
}
/*///////////////////////////////////////////////////////////////// */
/* Main keyboard code */

/*
	NOTE: keyboard hook function is currently ignored
*/


keyb_hook_t	pfn_keyb_hook = NULL;
dword_t		keyb_mode = KEYBOARD_FULLY_COOKED;

#define FLAG_LSHIFT		(0x01)
#define FLAG_LCTRL		(0x02)
#define FLAG_LALT		(0x04)
#define FLAG_SYSREQ		(0x08)
		/* System Request: Press CTRL+SysRq */

#define FLAG_RSHIFT		(0x10)
#define FLAG_RCTRL		(0x20)
#define FLAG_RALT		(0x40)
	/* 0x80 unused		 */

#define FLAG_CAPS		(0x0100)
#define FLAG_NUM		(0x0200)
#define FLAG_SCROLL		(0x0400)
	/* 0x0800 unused */

	/* 0x1000 unused */
	/* 0x2000 unused */
#define FLAG_DUAL		(0x4000)
#define FLAG_TREBLE		(0x8000)

#define KEY_DUAL		(0xE0)
#define KEY_TREBLE		(0x2A)
		/* follows a dual */

#define KEY_LSHIFT		(0x2A)
#define KEY_LCTRL		(0x1D)
#define KEY_LALT		(0x38)

/* basic keymap, should have 96 entries */
/* any key with a scan code of 0x60 or above will cause problems */
/* so we ignore these keys (see later) */
char_t keyMap[96] = {
	0, ASCII_ESC, '1', '2', '3', '4', '5', '6',
	'7', '8', '9', '0', '-', '=', ASCII_BS, ASCII_HT,
	'q', 'w', 'e', 'r', 't', 'y', 'u', 'i',
	'o', 'p', '[', ']', ASCII_LF, 0, 'a', 's',
	'd', 'f', 'g', 'h', 'j', 'k', 'l', ';',
	'\'', '`', 0, '#', 'z', 'x', 'c', 'v',
	'b', 'n', 'm', ',', '.', '/', 0, '*',
	0, ' ', 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, '7',
	'8', '9', '-', '4', '5', '6', '+', '1',
	'2', '3', '0', '.', 0, 0, '\\', 0,
	0, 0, 0, 0, 0, 0, 0, 0};

/* shifted keymap, again 96 entries */
char_t keyMapShift[96] = {
	0, ASCII_ESC, '!', '\"', '#', '$', '%', '^',
	'&', '*', '(', ')', '_', '+', ASCII_BS, ASCII_HT,
	'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I',
	'O', 'P', '{', '}', ASCII_LF, 0, 'A', 'S',
	'D', 'F', 'G', 'H', 'J', 'K', 'L', ':',
	'@', 0, 0, '~', 'Z', 'X', 'C', 'V',
	'B', 'N', 'M', '<', '>', '?', 0, '*',
	0, ' ', 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, '7',
	'8', '9', '-', '4', '5', '6', '+', '1',
	'2', '3', '0', '.', 0, 0, '|', 0,
	0, 0, 0, 0, 0, 0, 0, 0};

extern  void KeyboardInterrupt();
void InitKeyboardDevice();	/* at end of file */

/* interrupts must be disabled */
void InitKeyboard()
{
	VidWrite("Starting keyboard driver... ");
	KeyQueueInit();
	InitKeyboardDevice();
	InstallHandler(IDT_FIRST_EXTERNAL + 1, &KeyboardInterrupt);
	EnableIRQ(1, 1);
	VidWrite("Done\n");
}

void EnqueueHex(char_t c)
{
	byte_t high = (c >> 4) & 0x0F;
	byte_t low = c & 0x0F;
	char_t x;

	EnqueueKey('0');
	EnqueueKey('x');

	if (high < 0x0A)
		x = '0' + high;
	else
		x = 'A' + (high - 0x0A);

	EnqueueKey(x);

	if (low < 0x0A)
		x = '0' + low;
	else
		x = 'A' + (low - 0x0A);

	EnqueueKey(x);
}

/* called from the low level keyboard intrrupt handler */
extern 
void ProcessKeycode(word_t key)
{

	if ((key == KEY_DUAL) && !(keyState & FLAG_DUAL))
	{
		keyState |= FLAG_DUAL;
		return;
	}

	if ((key == KEY_TREBLE) && (keyState & FLAG_DUAL) && !(keyState & FLAG_TREBLE))
	{
		keyState &= ~FLAG_DUAL;
		keyState |= FLAG_TREBLE;
		return;
	}

	if ((key == (0x80 | KEY_TREBLE)) && (keyState & FLAG_TREBLE) && (keyState & FLAG_DUAL))
	{
		keyState &= ~FLAG_DUAL;
		keyState &= ~FLAG_TREBLE;
		return;
	}

	if (keyState & FLAG_TREBLE)		/* FLAG_DUAL will probably be set as well */
	{
		if (keyb_mode == KEYBOARD_SCANCODES)
		{
			if (!(key & 0x80))	/* don't print key-up scancodes */
			{
				EnqueueKey('=');
				EnqueueHex(key);
				EnqueueKey('\n');
			}
		}
		return;			/* ignore treble keys for now */
	}

	if (keyState & FLAG_DUAL)	/* !FLAG_TREBLE implied */
	{
		keyState &= ~FLAG_DUAL;
		if (keyb_mode == KEYBOARD_SCANCODES)
		{
			if (!(key & 0x80))
			{
				EnqueueKey('+');
				EnqueueHex(key);
				EnqueueKey('\n');
			}
		}
		return;			/* ignore dual-keys for now */
	}

	if (keyb_mode == KEYBOARD_SCANCODES)
	{
		if (!(key & 0x80))
		{
			EnqueueHex(key);
			EnqueueKey('\n');
		}
		return;
	}

	char_t x;

	if (key & 0x80)
	{
		key &= 0x7F;
		switch (key)
		{
			case KEY_LSHIFT:
				keyState &= ~FLAG_LSHIFT;
				break;
			case KEY_LCTRL:
				keyState &= ~FLAG_LCTRL;
				break;
			case KEY_LALT:
				keyState &= ~FLAG_LALT;
				break;
		}
		return;
	}

	switch (key)
	{
		case KEY_LSHIFT:
			keyState |= FLAG_LSHIFT;
			break;
		case KEY_LCTRL:
			keyState |= FLAG_LCTRL;
			break;
		case KEY_LALT:
			keyState |= FLAG_LALT;
			break;
		default:
			if (keyState & FLAG_LCTRL)	/* if we have LCTRL pressed, print the scancode */
			{
				PrintHex(key);
				return;
			}
			if (keyState & FLAG_LALT)
				return;			/* ignore LALT key combinations */
			if (key >= 0x60)
				return;			/* ignore keys outside of the character map */

			if (keyState & FLAG_LSHIFT)
				x = keyMapShift[key];
			else
				x = keyMap[key];

			EnqueueKey(x);		/* currently ignore errors */
	}

}

/*/////////////////////////////////////////////////////////////// */
/* keyboard device */

k_dev devKeyboard;

int_t KeybOpen(fileh f)
{
	return DEV_SUCCESS;
}

int_t KeybClose(fileh f)
{
	return DEV_SUCCESS;
}

int_t KeybRead(fileh f, char* ch)
{
	return DequeueKey(ch);
}

int_t KeybWrite(fileh f, char ch)
{
	return DEV_ERR_INVALID_REQUEST;
}

int_t KeybIoctl(fileh f, ulong_t request, void* arg)
{
	switch (request)
	{
		case KEYBOARD_SET_HOOK:
			pfn_keyb_hook = (keyb_hook_t)arg;
			break;
		case KEYBOARD_SET_MODE:
			{
				dword_t dw = (dword_t) arg;
				
				if ((dw != KEYBOARD_FULLY_COOKED) &&
				    (dw != KEYBOARD_RAW) &&
				    (dw != KEYBOARD_SCANCODES))
					return DEV_ERR_BAD_PARAM;
				
				keyb_mode = dw;
				break;
			}
		default:
			return DEV_ERR_INVALID_REQUEST;
	}
	
	return DEV_SUCCESS;
}

void InitKeyboardDevice()
{
	devKeyboard.flags = 0;
	devKeyboard.ulRecordSize = 0;
	
	devKeyboard.fnOpen = KeybOpen;
	devKeyboard.fnClose = KeybClose;
	devKeyboard.fnRead = KeybRead;
	devKeyboard.fnWrite = KeybWrite;
	devKeyboard.fnIoctl = KeybIoctl;
	
	RegisterDevice(KEYBOARD_DEVICE, &devKeyboard);
}
